#include <unistd.h>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <iostream>
#include <fstream>

//#include "inc/blas_demo.h"

#define __CL_ENABLE_EXCEPTIONS
#include <CL/cl.hpp>
#include "ocl_util.h"

using namespace cl;
using namespace std;


void print_matrix(int col_size, int row_size, float *mat)
{
  int i, j, k = 0;

  for ( i = 0; i < col_size; i++) {
    for ( j = 0; j < row_size; j++) {
      printf(" %4.4f", mat[k]);
      k++;
    }
    printf("\n");
  }
}

int main (int argc, char **argv)
{
   int   M = 2, N = 2, K = 3;
   float A[] = { 0.11, 0.12, 0.13,
                 0.21, 0.22, 0.23 };

   float B[] = { 1011, 1012,
                 1021, 1022,
                 1031, 1032 };

   float C[] = { 0.0, 0.0,
                 0.0, 0.0 };


   printf("Matrix A:\n");
   print_matrix(M,K,A);
   printf("\n");

   printf("Matrix B:\n");
   print_matrix(K,N,B);
   printf("\n");

   try
   {
     Context context(CL_DEVICE_TYPE_ACCELERATOR);
     std::vector<Device> devices = context.getInfo<CL_CONTEXT_DEVICES>();
     CommandQueue Q(context, devices[0], CL_QUEUE_PROFILING_ENABLE);
     /*------------------------------------------------------------------------
     * Read the offline compiled kernel module
     *-----------------------------------------------------------------------*/
     char *bin;
     int bin_length = ocl_read_binary("./kernel.out", bin);

     Program::Binaries   binary(1, std::make_pair(bin, bin_length));
     Program             program = Program(context, devices, binary);
     program.build(devices);

     delete [] bin;

     Buffer buf_matA(context, CL_MEM_READ_WRITE,M*K*sizeof(float));
     Buffer buf_matB(context, CL_MEM_READ_WRITE,K*N*sizeof(float));
     Buffer buf_matC(context, CL_MEM_READ_WRITE,M*N*sizeof(float));

     float *buf_matA_ptr = (float *)Q.enqueueMapBuffer(buf_matA, CL_TRUE, CL_MAP_WRITE, 0, M*K*sizeof(float));
     float *buf_matB_ptr = (float *)Q.enqueueMapBuffer(buf_matB, CL_TRUE, CL_MAP_WRITE, 0, K*N*sizeof(float));
     float *buf_matC_ptr = (float *)Q.enqueueMapBuffer(buf_matC, CL_TRUE, CL_MAP_WRITE, 0, M*N*sizeof(float));

     memcpy(buf_matA_ptr,A,M*K*sizeof(float));
     memcpy(buf_matB_ptr,B,K*N*sizeof(float));
     memcpy(buf_matC_ptr,C,M*N*sizeof(float));

     Q.enqueueUnmapMemObject(buf_matA, buf_matA_ptr);
     Q.enqueueUnmapMemObject(buf_matB, buf_matB_ptr);
     Q.enqueueUnmapMemObject(buf_matC, buf_matC_ptr);

     Event ev;
     /*---------------------------------------------------------------------
     * Create a fftlib plan in msmc for subsequent use
     *--------------------------------------------------------------------*/
     Kernel BLAS_sgemm      (program, "ocl_BLAS_sgemm");

     BLAS_sgemm.setArg(0,M);
     BLAS_sgemm.setArg(1,N);
     BLAS_sgemm.setArg(2,K);
     BLAS_sgemm.setArg(3,buf_matA);
     BLAS_sgemm.setArg(4,buf_matB);
     BLAS_sgemm.setArg(5,buf_matC);

     Q.enqueueTask(BLAS_sgemm,0,&ev);

     ev.wait();

     buf_matC_ptr = (float *)Q.enqueueMapBuffer(buf_matC, CL_TRUE, CL_MAP_READ, 0, M*N*sizeof(float));
     printf("Matrix C = A*B:\n");
     print_matrix(M,N,buf_matC_ptr);
   }
   catch (Error err)
   { cerr << "ERROR: " << err.what() << "(" << err.err() << ")" << endl; }


   return 0;
}


